module AppWithUndo exposing (addUndoButton)

import Browser
import Html exposing (Html, button, div, text)
import Html.Attributes
import Html.Events exposing (onClick)



{--

msg stands for the Message type of the inner model; 
    it varies from application to application  so it's a type variable 

mdl stands for the Model type of the inner model 


message stands for an object of type msg; that is, it's always an inner message 
model stands for an object of type mdl, the inner model


wrapMessage is an object of type (WrapMessage msg)

wrapModel is an object of type (WrapModel msg)
--}


type alias WrapModel mdl msg =
    List ( mdl, Maybe msg )



-- past and present


type WrapMessage msg
    = Undo
    | InnerMessage msg


type alias ElementArgs flags mdl msg =
    { init : flags -> ( mdl, Cmd msg )
    , subscriptions : mdl -> Sub msg
    , update : msg -> mdl -> ( mdl, Cmd msg )
    , view : mdl -> Html msg
    }


wrapInit :
    (flags -> ( mdl, Cmd msg ))
    -> (flags -> ( WrapModel mdl msg, Cmd (WrapMessage msg) ))
wrapInit innerInit =
    let
        outerInit : flags -> ( WrapModel mdl msg, Cmd (WrapMessage msg) )
        outerInit theFlags =
            let
                ( initialInnerModel, initialInnerCommand ) =
                    innerInit theFlags

                initialOuterModel =
                    [ ( initialInnerModel, Nothing ) ]

                initialOuterCommand =
                    Cmd.map InnerMessage initialInnerCommand
            in
            ( initialOuterModel, initialOuterCommand )
    in
    outerInit



{--wrapUpdate recordMessage innerUpdate

innerUpdate : the update function of the inner model 
recordMessage : msg -> Bool, tells us whether we should record 
the inner message or whether we should ignore it, lest they flood 
our history, like mouse move event messages, for example 

--}


wrapUpdate :
    (msg -> Bool)
    -> (msg -> mdl -> ( mdl, Cmd msg ))
    -> (WrapMessage msg -> WrapModel mdl msg -> ( WrapModel mdl msg, Cmd (WrapMessage msg) ))
wrapUpdate recordMessage innerUpdate =
    let
        outerUpdate : WrapMessage msg -> WrapModel mdl msg -> ( WrapModel mdl msg, Cmd (WrapMessage msg) )
        outerUpdate wrapMessage wrapModel =
            case ( wrapMessage, wrapModel ) of
                ( _, [] ) ->
                    ( [], Cmd.none )

                -- not going to happen
                ( Undo, ( currentModel, lastMessage ) :: (previousModel :: history) ) ->
                    let
                        xyz =
                            Debug.log "undoing message" (Debug.toString lastMessage)

                        xxx =
                            Debug.log "history length is " (String.fromInt (List.length wrapModel))
                    in
                    ( previousModel :: history, Cmd.none )

                ( Undo, [ currentModel ] ) ->
                    let
                        xyz =
                            Debug.log "undo clicked" ""
                    in
                    ( [ currentModel ], Cmd.none )

                -- undo the last mod step
                ( InnerMessage message, ( currentModel, lastMessage ) :: history ) ->
                    let
                        ( newModel, newCmd ) =
                            innerUpdate message currentModel
                    in
                    if recordMessage message then
                        let
                            asdfjkla =
                                Debug.log "recording message" (Debug.toString message)
                        in
                        ( ( newModel, Just message ) :: wrapModel, Cmd.map InnerMessage newCmd )

                    else
                        ( ( newModel, lastMessage ) :: history, Cmd.map InnerMessage newCmd )
    in
    outerUpdate


wrapView :
    (mdl -> Html msg)
    -> (WrapModel mdl msg -> Html (WrapMessage msg))
wrapView innerView =
    let
        outerView : WrapModel mdl msg -> Html.Html (WrapMessage msg)
        outerView wrapModel =
            case wrapModel of
                [] ->
                    div [] [ text "blabla" ]

                -- this will never happen
                ( currentModel, lastMessage ) :: history ->
                    let
                        innerHtmlObjects : Html msg
                        innerHtmlObjects =
                            innerView currentModel
                    in
                    div []
                        [ Html.map InnerMessage innerHtmlObjects
                        , Html.hr [ Html.Attributes.style "border" "1pt solid black" ] []
                        , div [] [ button [ onClick Undo ] [ text "undo" ] ]
                        ]
    in
    outerView



-- will ignore the input parameter, i.e., the model
-- because I haven't used this so far


wrapSubscriptions :
    (mdl -> Sub msg)
    -> (WrapModel mdl msg -> Sub (WrapMessage msg))
wrapSubscriptions innerSub =
    let
        outerSubscription : WrapModel mdl msg -> Sub (WrapMessage msg)
        outerSubscription outerModel =
            case outerModel of
                [] ->
                    Sub.none

                ( innerModel, lastMessage ) :: history ->
                    let
                        sub =
                            innerSub innerModel
                    in
                    Sub.map InnerMessage sub
    in
    outerSubscription


addUndoButton : (msg -> Bool) -> ElementArgs flags mdl msg -> ElementArgs flags (WrapModel mdl msg) (WrapMessage msg)
addUndoButton recordMessage elementArgs =
    { init = wrapInit elementArgs.init
    , update = wrapUpdate recordMessage elementArgs.update
    , view = wrapView elementArgs.view
    , subscriptions = wrapSubscriptions elementArgs.subscriptions
    }



{--
element : ElementArgs flags model msg -> Program flags (WrapModel model) (WrapMessage msg)
element args =
    Browser.element
        { init = wrapInit args.init
        , update = wrapUpdate args.update
        , view = wrapView args.view
        , subscriptions = \_ -> Sub.none -- to be done; correct only if inner program has no subscriptions
        }
--}
