module AuxiliaryStuff exposing (..)


firstWith : (a -> Bool) -> List a -> Maybe a
firstWith condition list =
    case list of
        [] ->
            Nothing

        x :: rest ->
            if condition x then
                Just x

            else
                firstWith condition rest



{--


    Tries to apply f to each element of x and stops after the first 
    successful application; unsuccessful elements remain unchanged, the remaining elements will be 
    unchanged. It also returns a flag to indicate whether any element has been replaced at all

> replaceFirstSuccessful halveIfPossible [1,3,3,4,5]
[1,3,3,2,5] : List Int    


    Takes a list [x1, ..., xn] and returns 
    [x1, ..., x_{i-1}, f x_i, x_{i+1}, x_n]
    where x_i is the first element for which f is applied successfully. 



--}


replaceFirstSuccessful : (a -> Maybe a) -> List a -> ( Bool, List a )
replaceFirstSuccessful f list =
    case list of
        [] ->
            ( False, [] )

        x :: rest ->
            case f x of
                -- f on x not yet successful
                Nothing ->
                    let
                        ( success, restReplaced ) =
                            replaceFirstSuccessful f rest
                    in
                    ( success, x :: restReplaced )

                Just newX ->
                    ( True, newX :: rest )


halveIfPossible : Int -> Maybe Int
halveIfPossible x =
    if modBy 2 x == 0 then
        Just (x // 2)

    else
        Nothing


exampleList =
    [ 1, 3, 3, 5, 6, 7, 8, 9 ]


exampleListReplaced =
    replaceFirstSuccessful halveIfPossible exampleList


smallestNotIn : List comparable -> List comparable -> Maybe comparable
smallestNotIn good bad =
    List.minimum (List.filter (\x -> not (List.member x bad)) good)


cartesianProduct : List String -> List String -> List String
cartesianProduct list1 list2 =
    case list1 of
        [] ->
            []

        a :: rest1 ->
            List.map (\b -> a ++ b) list2 ++ cartesianProduct rest1 list2


alphabet =
    List.map String.fromChar (String.toList "abcdefghijklmnopqrstuvwxyz")


freshVariable : List String -> String
freshVariable forbidden =
    freshVariableRec alphabet forbidden


freshVariableRec candidates forbidden =
    case smallestNotIn candidates forbidden of
        Just s ->
            s

        Nothing ->
            freshVariableRec (cartesianProduct alphabet candidates) forbidden


divideIntervalByWeights : ( Float, Float ) -> List Float -> List ( Float, Float )
divideIntervalByWeights ( left, right ) weights =
    case weights of
        [] ->
            []

        -- should never happen
        w1 :: restWeights ->
            let
                -- accumulated weights: This has quadratic complexity but it's succinct and obscure...
                accumWeights : List Float
                accumWeights =
                    List.foldr (\x list -> x :: List.map ((+) x) list) [] weights

                totalWeight =
                    List.foldl (+) 0 weights

                accumWeightToPosition x =
                    left + x / totalWeight * (right - left)

                positions =
                    List.map accumWeightToPosition accumWeights

                p1 =
                    accumWeightToPosition w1

                -- now convert [0, z1, z2, z3, ..., zk ]
                -- into k pieces
                -- [ (0,z1), (z1,z2), (z2, z3), ...]
                foldingfunction : Float -> List ( Float, Float ) -> List ( Float, Float )
                foldingfunction x list =
                    case list of
                        [] ->
                            [ ( p1, left ) ]

                        ( y, z ) :: restlist ->
                            ( x, y ) :: ( y, z ) :: restlist

                intervals : List ( Float, Float )
                intervals =
                    List.reverse (List.map (\( x, y ) -> ( y, x )) (List.foldl foldingfunction [] positions))
            in
            intervals



{--
divideInterval : ( Float, Float ) -> Int -> List HorizontalRegion
divideInterval ( left, right ) pieces =
    case pieces of
        0 ->
            []

        1 ->
            [ HorizontalRegion { left = left, center = (left + right) / 2.0, right = right } ]

        _ ->
            let
                subinterval_width =
                    (right - left) / toFloat pieces
            in
            List.map
                (\i ->
                    let
                        leftHere =
                            left + toFloat i * subinterval_width

                        centerHere =
                            leftHere + subinterval_width * 0.5

                        rightHere =
                            leftHere + subinterval_width
                    in
                    HorizontalRegion { left = leftHere, center = centerHere, right = rightHere }
                )
                (List.range 0 (pieces - 1))

--}
{--
 
--}


map2lists : (a -> b -> c) -> List a -> List b -> List c
map2lists whichFunction firstList secondList =
    case ( firstList, secondList ) of
        ( [], _ ) ->
            []

        ( _, [] ) ->
            []

        ( x :: rest1, y :: rest2 ) ->
            whichFunction x y :: map2lists whichFunction rest1 rest2


applyFunctionsToArguments : List (a -> b) -> List a -> List b
applyFunctionsToArguments listOfFunctions listOfArguments =
    case ( listOfFunctions, listOfArguments ) of
        ( [], _ ) ->
            []

        ( _, [] ) ->
            []

        ( f :: restListOfFunctions, x :: restListOfArguments ) ->
            f x :: applyFunctionsToArguments restListOfFunctions restListOfArguments


getNth : Int -> List a -> Maybe a
getNth n list =
    case ( n, list ) of
        ( _, [] ) ->
            Nothing

        ( 0, x :: _ ) ->
            Just x

        ( _, x :: rest ) ->
            getNth (n - 1) rest


getEvenIndices : List a -> List a
getEvenIndices list =
    case list of
        [] ->
            []

        x :: [] ->
            [ x ]

        x :: y :: rest ->
            x :: getEvenIndices rest


listMaybeToMaybeList : List (Maybe a) -> Maybe (List a)
listMaybeToMaybeList list =
    case list of
        [] ->
            Just []

        (Just x) :: restlist ->
            case listMaybeToMaybeList restlist of
                Just resultList ->
                    Just (x :: resultList)

                _ ->
                    Nothing

        _ ->
            Nothing


removeDuplicates : List a -> List a
removeDuplicates list =
    let
        addOne : a -> List a -> List a
        addOne x rest =
            if List.member x rest then
                rest

            else
                x :: rest
    in
    List.foldr addOne [] list


mergeResultList : List (Result String a) -> Result String (List a)
mergeResultList list =
    case list of
        [] ->
            Result.Ok []

        (Result.Ok x) :: rest ->
            case mergeResultList rest of
                Result.Ok restMerged ->
                    Result.Ok (x :: restMerged)

                Result.Err msg ->
                    Result.Err msg

        (Result.Err s) :: rest ->
            case mergeResultList rest of
                Result.Ok restMerged ->
                    Result.Err s

                Result.Err msg ->
                    Result.Err (s ++ "; " ++ msg)


findKey : List ( a, b ) -> a -> Maybe b
findKey list key =
    case list of
        [] ->
            Nothing

        ( x, y ) :: rest ->
            if x == key then
                Just y

            else
                findKey rest key
