module Tree exposing (..)

import AuxiliaryStuff


type Tree a
    = Tree { data : a, children : List (Tree a) }


getData : Tree a -> a
getData (Tree tree) =
    tree.data


getChildren : Tree a -> List (Tree a)
getChildren (Tree tree) =
    tree.children


getChild : Maybe (Tree a) -> Int -> Maybe (Tree a)
getChild maybeTree n =
    case maybeTree of
        Just (Tree tree) ->
            getNthElement tree.children n

        Nothing ->
            Nothing


isLeaf : Tree a -> Bool
isLeaf (Tree tree) =
    tree.children == []


getDataAt : Maybe (Tree a) -> List Int -> Maybe a
getDataAt maybeTree path =
    let
        maybeNode =
            getNode maybeTree path
    in
    case maybeNode of
        Nothing ->
            Nothing

        Just node ->
            Just (getData node)


getNode : Maybe (Tree a) -> List Int -> Maybe (Tree a)
getNode maybeTree path =
    case maybeTree of
        Nothing ->
            Nothing

        Just (Tree tree) ->
            case path of
                [] ->
                    maybeTree

                i :: restpath ->
                    getNode (getChild maybeTree i) restpath


getDataAlongPath : Maybe (Tree a) -> List Int -> Maybe (List a)
getDataAlongPath maybeTree path =
    case maybeTree of
        Nothing ->
            Nothing

        Just (Tree tree) ->
            case path of
                [] ->
                    Just [ tree.data ]

                i :: restpath ->
                    let
                        dataAlongRestPath =
                            getDataAlongPath (getChild maybeTree i) restpath
                    in
                    case dataAlongRestPath of
                        Nothing ->
                            Nothing

                        Just theRealDataAlongRestPath ->
                            Just (tree.data :: theRealDataAlongRestPath)


leaf : a -> Tree a
leaf x =
    Tree { data = x, children = [] }


map : (a -> b) -> Tree a -> Tree b
map f (Tree tree) =
    Tree
        { data = f tree.data
        , children = List.map (map f) tree.children
        }


mapWithContext : (Tree a -> b) -> Tree a -> Tree b
mapWithContext f (Tree tree) =
    Tree { data = f (Tree tree), children = List.map (mapWithContext f) tree.children }


filter : (a -> Bool) -> Tree a -> Tree a
filter condition (Tree tree) =
    let
        filteredChildren : List (Tree a)
        filteredChildren =
            List.filter (\child -> condition (getData child)) tree.children

        recursivelyFilteredChildren =
            List.map (filter condition) filteredChildren
    in
    Tree { data = tree.data, children = recursivelyFilteredChildren }


toString : (a -> String) -> Tree a -> String
toString dataToString tree =
    toStringRec dataToString "" tree


toStringRec : (a -> String) -> String -> Tree a -> String
toStringRec dataToString prefix (Tree tree) =
    let
        toStringChildren : Tree a -> String
        toStringChildren t =
            toStringRec dataToString (" | " ++ prefix) t ++ "\n"

        childrenStrings =
            List.map toStringChildren tree.children
    in
    prefix ++ dataToString tree.data ++ "\n" ++ List.foldr (++) "" childrenStrings



-- if the subtree of a node has only whitespace at its leaves, it will be deleted


cleanWhitespace : Tree String -> Tree String
cleanWhitespace tree =
    Tuple.first (cleanWhitespaceRec tree)



-- the second argument tells us whether this node is to be considerede 'nontrivial'
-- a tree is nontrivial if
-- it used to be a leaf in the original tree and its value is not only whitespace
-- it used to be an inner node in the original tree and now still has children


cleanWhitespaceRec : Tree String -> ( Tree String, Bool )
cleanWhitespaceRec (Tree tree) =
    case tree.children of
        [] ->
            if String.trim tree.data == "" then
                ( Tree { data = "", children = [] }, False )
                -- a trivial node

            else
                ( Tree tree, True )

        -- it's a leaf with text, so it's nontrivial
        _ ->
            -- this is not a leaf; so recursively clean up the kids
            let
                resultOfRec : List ( Tree String, Bool )
                resultOfRec =
                    List.map cleanWhitespaceRec tree.children

                nontrivialResults =
                    List.filter Tuple.second resultOfRec

                newChildren =
                    List.map Tuple.first nontrivialResults
            in
            -- it's an inner node; at least used to be;
            -- so to be nontrivial it needs children
            ( Tree { data = tree.data, children = newChildren }, newChildren /= [] )


getNthElement : List a -> Int -> Maybe a
getNthElement list n =
    case ( list, n ) of
        ( [], _ ) ->
            Nothing

        ( x :: restlist, 0 ) ->
            Just x

        ( _ :: restlist, _ ) ->
            getNthElement restlist (n - 1)


setNthElement : List a -> Int -> a -> Maybe (List a)
setNthElement list n new_a =
    case ( list, n ) of
        ( [], _ ) ->
            Nothing

        ( x :: restlist, 0 ) ->
            Just (new_a :: restlist)

        ( x :: restlist, _ ) ->
            case setNthElement restlist (n - 1) new_a of
                Nothing ->
                    Nothing

                Just newRestList ->
                    Just (x :: newRestList)



{--
    Finds the path to the first node to match a condition; 
    depth-first search, left children first 
    
    Returns a pair: the node itself and the path (in top-down order) from the root to the child;

    so a path [3,0,1] means 

    - take child 3 of the root 
    - take child 0 thereof
    - take child 1 thereof and this is your result 

--}


firstWith : (Tree a -> Bool) -> Tree a -> Maybe (List Int)
firstWith condition (Tree tree) =
    if condition (Tree tree) then
        Just []

    else
        firstWithAmongListOfTrees condition 0 tree.children


firstWithAmongListOfTrees : (Tree a -> Bool) -> Int -> List (Tree a) -> Maybe (List Int)
firstWithAmongListOfTrees condition indexOfChild listOfTrees =
    case listOfTrees of
        [] ->
            Nothing

        tree :: restOfTrees ->
            case firstWith condition tree of
                Nothing ->
                    firstWithAmongListOfTrees condition (indexOfChild + 1) restOfTrees

                Just path ->
                    Just (indexOfChild :: path)



{--
    Finds the first node u (top to bottom, left to right) that matches the condition
    and replaces u's children by the list provided.

    Returns Just (the new tree) if a node satisfying the condition has been found, 
        otherwise returns Nothing
--}


findFirstAndReplaceChildren : (Tree a -> Bool) -> List (Tree a) -> Tree a -> Maybe (Tree a)
findFirstAndReplaceChildren condition newValues (Tree tree) =
    if condition (Tree tree) then
        Just (Tree { tree | children = newValues })

    else if tree.children == [] then
        Nothing

    else
        -- root does not match condition but it has some children
        let
            ( success, newChildren ) =
                AuxiliaryStuff.replaceFirstSuccessful
                    (findFirstAndReplaceChildren condition newValues)
                    tree.children
        in
        if success then
            Just
                (Tree
                    { tree
                        | children =
                            newChildren
                    }
                )

        else
            Nothing


exampleTree =
    Tree { data = 3, children = [ leaf 1, leaf 2, Tree { data = 3, children = [ leaf 5, leaf 4 ] }, leaf 8 ] }


exampleTreeReplaced =
    findFirstAndReplaceChildren (\x -> modBy 2 (getData x) == 0) [ leaf -1, leaf -1 ] exampleTree


replaceSubtreeRec : Tree a -> List Int -> Tree a -> Maybe (Tree a)
replaceSubtreeRec (Tree tree) path newSubtree =
    case path of
        [] ->
            Just newSubtree

        -- replace the root itself
        n :: restpath ->
            case getNthElement tree.children n of
                Nothing ->
                    Nothing

                -- not even subtree could be located
                Just childTree ->
                    case replaceSubtreeRec childTree restpath newSubtree of
                        Nothing ->
                            Nothing

                        -- subtree replacement failed in recursion
                        Just newChild ->
                            case setNthElement tree.children n newChild of
                                Just newListOfChildren ->
                                    Just (Tree { data = tree.data, children = newListOfChildren })

                                Nothing ->
                                    Nothing



{--
tree1 =
    Tree { data = "1", children = [ leaf "2", leaf "3", leaf "4" ] }


tree2 =
    replaceSubtreeRec tree1 [ 1 ] (Tree { data = "bla", children = [ leaf "zwei" ] })


tree2AsString =
    case tree2 of
        Nothing ->
            "Nothing"

        Just t ->
            toString t
--}


decayFactor =
    0.5


count : Tree a -> Float
count tree =
    countNodesExponentialDecay 1 tree


countNodesExponentialDecay : Float -> Tree a -> Float
countNodesExponentialDecay decay (Tree tree) =
    1
        + decay
        * List.foldl
            (+)
            0
            (List.map (countNodesExponentialDecay decay) tree.children)
