module TreeToSvg exposing (..)

import AuxiliaryStuff
import Html.Attributes
import Svg exposing (Svg)
import Svg.Attributes
import SvgAuxiliary exposing (Box)
import Tree exposing (..)



{--
    metaInfo: info how to draw the nodes, like color of the dots, whether node is "valid", whatever 


--}


type alias DrawTreeMetaInfo =
    { dotFillColor : String
    , dotStrokeColor : String
    , lineStrokeColor : String
    , strokeWidth : Float
    , zoom : Float
    }


type alias DrawTreeData msg =
    { metaInfo : DrawTreeMetaInfo
    , lineAttributes : List (Svg.Attribute msg)
    , dotAttributes : List (Svg.Attribute msg)
    , svgLabelString : String
    }



{--draw the tree with one color scheme and no event handling etc. --}


drawTreePlain : Bool -> Float -> Box -> Tree String -> List (Svg msg)
drawTreePlain drawRoot zoom rootBox tree =
    let
        metaInfo : DrawTreeMetaInfo
        metaInfo =
            { dotFillColor = "cornsilk"
            , dotStrokeColor = SvgAuxiliary.originalColorScheme.dotStroke
            , lineStrokeColor = SvgAuxiliary.originalColorScheme.lines
            , strokeWidth = SvgAuxiliary.strokeWidth
            , zoom = zoom
            }

        useAndUpdateDrawInfo : DrawTreeMetaInfo -> Tree String -> List Int -> DrawTreeData msg
        useAndUpdateDrawInfo someInfo (Tree someTree) somePath =
            let
                labelString =
                    someTree.data

                newDrawTreeData : DrawTreeData msg
                newDrawTreeData =
                    { metaInfo = metaInfo
                    , lineAttributes = []
                    , dotAttributes = []
                    , svgLabelString = labelString
                    }
            in
            newDrawTreeData
    in
    drawTreeWithOrWithoutRoot drawRoot
        useAndUpdateDrawInfo
        metaInfo
        { rootBox | top = rootBox.top }
        tree
        []


drawTree :
    (DrawTreeMetaInfo -> Tree dataType -> List Int -> DrawTreeData msg)
    -> DrawTreeMetaInfo
    -> Box
    -> Tree dataType
    -> List Int
    -> List (Svg msg)
drawTree useAndUpdateDrawInfo oldMetaInfo box (Tree tree) pathToHere =
    drawTreeWithOrWithoutRoot True useAndUpdateDrawInfo oldMetaInfo box (Tree tree) pathToHere



{--
useAndUpdateDrawInfo produces the new draw info to be passed to the children 
and the attributes for the lines and for the dots.

Attributes could for example be event listeners. 

Color and stroke width attributes do not need to be set from the outside, 
    they will be set using the DrawTreeMetaInfo provided 
--}


drawTreeWithOrWithoutRoot :
    Bool
    -> (DrawTreeMetaInfo -> Tree dataType -> List Int -> DrawTreeData msg)
    -> DrawTreeMetaInfo
    -> Box
    -> Tree dataType
    -> List Int
    -> List (Svg msg)
drawTreeWithOrWithoutRoot drawRoot useAndUpdateDrawInfo oldMetaInfo box (Tree tree) pathToHere =
    let
        { metaInfo, lineAttributes, dotAttributes, svgLabelString } =
            useAndUpdateDrawInfo oldMetaInfo (Tree tree) pathToHere

        root_x =
            (box.right + box.left) / 2

        root_y =
            box.top

        box_width =
            box.right - box.left

        shrinkFactor =
            box_width / (SvgAuxiliary.rootBox.right - SvgAuxiliary.rootBox.left)

        root =
            ( root_x, root_y )

        edge_y_width =
            (1 - SvgAuxiliary.y_decay_factor) * (box.bottom - box.top)

        child_y =
            box.top + edge_y_width

        listOfTreeWeights =
            List.map (Tree.countNodesExponentialDecay Tree.decayFactor) tree.children

        listOfChildrenIntervals : List ( Float, Float )
        listOfChildrenIntervals =
            AuxiliaryStuff.divideIntervalByWeights
                ( box.left, box.right )
                listOfTreeWeights

        listOfChildrenXcoordinates : List Float
        listOfChildrenXcoordinates =
            List.map (\( x1, x2 ) -> (x1 + x2) / 2) listOfChildrenIntervals

        listOfChildrenBoxes : List Box
        listOfChildrenBoxes =
            let
                intervalToBox : ( Float, Float ) -> Box
                intervalToBox ( x1, x2 ) =
                    let
                        boxWidth =
                            x2 - x1

                        {--boxHeight =
                            boxWidth / root_aspect_ratio --}
                    in
                    { left = x1
                    , right = x2
                    , top = child_y
                    , bottom = SvgAuxiliary.rootBox.bottom

                    --, bottom = child_y + boxHeight
                    }
            in
            List.map intervalToBox listOfChildrenIntervals

        linesAsSvg : List (Svg msg)
        linesAsSvg =
            if drawRoot then
                List.map
                    (SvgAuxiliary.createLine
                        [ Svg.Attributes.stroke metaInfo.lineStrokeColor
                        , Svg.Attributes.strokeWidth (String.fromFloat (metaInfo.strokeWidth * metaInfo.zoom))
                        ]
                        ( root_x, root_y )
                    )
                    (List.map (\x -> ( x, child_y )) listOfChildrenXcoordinates)

            else
                []

        indicesOfChildren : List Int
        indicesOfChildren =
            List.range 0 (List.length tree.children - 1)

        pathsToEachChild : List (List Int)
        pathsToEachChild =
            List.map
                (\x -> x :: pathToHere)
                indicesOfChildren

        {--
            so basically, we want to do 
            result = []
            for (int i = 0; i < tree.children.length; i++) {
                result.append(
                  drawTree( model newColorScheme listOfChildrenBoxes[i] tree.children[i] pathsToEachChild[i])
                )
            }
            but since we don't have for-loops, we first batch apply the Curryed function
            drawTree model newColorScheme 
            via List.map to listOfChildrenBoxes.

            This gives us a list of functions, which we apply point-by-point to tree.children

            and so on.

        --}
        -- drawTree useAndUpdateDrawInfo drawInfo box  (Tree tree) pathToHere
        -- from now on always draw the root
        resultsOfDrawTree1 : List (Tree dataType -> List Int -> List (Svg msg))
        resultsOfDrawTree1 =
            List.map (drawTreeWithOrWithoutRoot True useAndUpdateDrawInfo metaInfo) listOfChildrenBoxes

        resultsOfDrawTree2 : List (List Int -> List (Svg msg))
        resultsOfDrawTree2 =
            AuxiliaryStuff.applyFunctionsToArguments resultsOfDrawTree1 tree.children

        resultsOfDrawTree3 : List (List (Svg msg))
        resultsOfDrawTree3 =
            AuxiliaryStuff.applyFunctionsToArguments resultsOfDrawTree2 pathsToEachChild

        subTreesAsSvg : List (Svg msg)
        subTreesAsSvg =
            List.foldl (++) [] resultsOfDrawTree3

        rootDotAndLabel : List (Svg msg)
        rootDotAndLabel =
            if drawRoot then
                [ SvgAuxiliary.createDot
                    ([ Svg.Attributes.stroke metaInfo.dotStrokeColor
                     , Svg.Attributes.fill metaInfo.dotFillColor
                     , Svg.Attributes.strokeWidth (String.fromFloat (metaInfo.strokeWidth * metaInfo.zoom))
                     ]
                        ++ dotAttributes
                    )
                    ( root_x, root_y )
                    (SvgAuxiliary.circleRadius * shrinkFactor ^ 0.2 * metaInfo.zoom ^ 0.6)
                , SvgAuxiliary.createSvgLabel
                    root
                    svgLabelString
                    (24 * shrinkFactor ^ 0.2 * metaInfo.zoom ^ 0.6)
                ]

            else
                []
    in
    linesAsSvg
        ++ subTreesAsSvg
        ++ rootDotAndLabel
